package com.bawt.update;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.*;
import org.apache.bcel.Constants;

import java.util.HashMap;
import java.util.Iterator;

import com.bawt.hook.*;

@SuppressWarnings({"deprecation"})
public class ClientPatcher {
    public static void main(String args[]) throws Exception {
        new ClientPatcher(ClientPackage.unpack());
    }

    private ClassGen classes[];
    private HashMap<String, String> classnames;
    private HashMap<String, String> fieldnames;

    public void simplifyShifts() {
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            for (Method m : cg.getMethods()) {
                if (!m.isAbstract()) {
                    MethodGen mg = new MethodGen(m, cg.getClassName(), cpg);
                    InstructionList il = mg.getInstructionList();
                    InstructionFinder ifinder = new InstructionFinder(il);
                    Iterator<InstructionHandle[]> it = ifinder.search("LDC (LSHR|LSHL|LUSHR|ISHL|ISHR|IUSHR)");
                    while (it.hasNext()) {
                        InstructionHandle[] ih = it.next();
                        InstructionHandle first = ih[0];
                        Instruction shift = ih[1].getInstruction();

                        LDC ldc = (LDC) first.getInstruction();
                        int value = ((Integer) ldc.getValue(cpg)).intValue();

                        if (shift instanceof LSHR || shift instanceof LSHL || shift instanceof LUSHR) {
                            value &= 63;
                        } else {
                            value &= 31;
                        }

                        first.setInstruction(new BIPUSH((byte) value));
                    }

                    mg.setInstructionList(il);
                    mg.setMaxLocals();
                    mg.setMaxStack();
                    cg.replaceMethod(m, mg.getMethod());
                }
            }
        }
    }

    public void simplifyConditionalsEQ() throws Exception {
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            for (Method m : cg.getMethods()) {
                if (!m.isAbstract()) {
                    MethodGen mg = new MethodGen(m, cg.getClassName(), cpg);
                    InstructionList il = mg.getInstructionList();
                    InstructionFinder ifinder = new InstructionFinder(il);
                    Iterator<InstructionHandle[]> it = ifinder.search("ICONST_M1 IXOR (IFEQ|IFNE|IFLT|IFLE|IFGT|IFGE)");
                    if (it.hasNext()) {
                        InstructionHandle[] ih = it.next();
                        il.delete(ih[1]);
                        Instruction ii = ih[2].getInstruction();
                        if (ii instanceof IFEQ) {
                            ih[2].setInstruction(new IF_ICMPEQ(((IfInstruction) ii).getTarget()));
                        } else if (ii instanceof IFNE) {
                            ih[2].setInstruction(new IF_ICMPNE(((IfInstruction) ii).getTarget()));
                        } else if (ii instanceof IFGE) {
                            il.delete(ih[0]);
                            ih[2].setInstruction(new IFLT(((IfInstruction) ii).getTarget()));
                        } else if (ii instanceof IFGT) {
                            ih[2].setInstruction(new IF_ICMPLT(((IfInstruction) ii).getTarget()));
                        } else if (ii instanceof IFLE) {
                            ih[2].setInstruction(new IF_ICMPGE(((IfInstruction) ii).getTarget()));
                        } else if (ii instanceof IFLT) {
                            il.delete(ih[0]);
                            ih[2].setInstruction(new IFGE(((IfInstruction) ii).getTarget()));
                        }
                    }

                    mg.setInstructionList(il);
                    mg.setMaxLocals();
                    mg.setMaxStack();
                    cg.replaceMethod(m, mg.getMethod());
                }
            }
        }
    }

    public ClientPatcher(ClassGen cg[]) throws Exception {
        classnames = new HashMap<String, String>();
        fieldnames = new HashMap<String, String>();
        classes = cg;

//        simplifyShifts();
//        simplifyConditionalsEQ();

        ClassGen client = findClass("client");
        client.addInterface("com.bawt.hook.Client");

        Method main = client.containsMethod("main", "([Ljava/lang/String;)V");
        MethodGen mg = new MethodGen(main, "client", client.getConstantPool());
        Instruction instructions[] = mg.getInstructionList().getInstructions();
        for (Instruction i : instructions) {
            if (i instanceof ConstantPushInstruction) {
                ConstantPushInstruction cpi = (ConstantPushInstruction) i;
                int value = cpi.getValue().intValue();
                if (value != 503 && value != 765 && (value >= 445 && value < 500)) {
                    System.out.println("Runescape Build #" + value);
                    ClientPackage.backup(value);
                    break;
                }
            }
        }

        // TODO Hack up the "applet" class to override FocusListener, WindowListener, and getGraphics.   
//        hackApplet();
        hackMouse();
        hackKeyBoard();
        hackCanvas();
        hackArrayCopy();

        long begin = System.currentTimeMillis();

        if (!hookUserInput()) {
            throw new Exception("Problem hooking user's input.");
        }

        if (!hookCanvas()) {
            throw new Exception("Problem hooking Canvas.");
        }

        if (!hookStringWrapper()) {
            throw new Exception("Problem hooking StringWrapper.");
        }

        if (!hookUserPass()) {
            throw new Exception("Problem hooking Username and Password.");
        }

        if (!hookPlayerArrays()) {
            throw new Exception("Problem hooking Player Arrays.");
        }

        if (!hookPlayerCharacter()) {
            throw new Exception("Error hooking players and character.");
        }

        if (!hookMonsterArrays()) {
            throw new Exception("Problem hooking Monster Arrays.");
        }

        if (!hookMonsterComposite()) {
            throw new Exception("Error hooking monster and composite.");
        }

        if (!hookMonsterID()) {
            throw new Exception("Problem locating NPC ID!");
        }

        if (!hookLevels()) {
            throw new Exception("Problem hooking levels.");
        }

        if (!hookMyPlayer()) {
            throw new Exception("Problem hooking our player.");
        }

        if (!hookChatArrays()) {
            throw new Exception("Problem hooking chat arrays.");
        }

        if (!hookChatStates()) {
            throw new Exception("Problem hooking chat states.");
        }

        if (!hookWorldCoords()) {
            throw new Exception("Problem hooking world coordinates.");
        }

        if (!hookCharacterHpInfo()) {
            throw new Exception("Problem hooking hp information.");
        }

        if (!hookCharacterCoordinates()) {
            throw new Exception("Problem hooking character coordinates.");
        }

        if (!hookExpMaxes()) {
            throw new Exception("Problem hooking exp max array.");
        }

        if (!hookSkills()) {
            throw new Exception("Problem hooking skill arrays.");
        }

        hookMenu();

        System.out.println("Updated in " + (System.currentTimeMillis() - begin) + "ms.");
    }

    public void hookMenu() {
        ClassGen client = findClass("client");
        Member choose = findString("Choose Option");
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupFieldref(choose.classname, choose.name, "L" + choose.signature + ";");
            if (lookup != -1) {
                String classname = c.getClassName();
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        if (m.getName().indexOf("init") == -1) {
                            MethodGen mg = new MethodGen(m, classname, cpg);
                            InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                            if (hook.hookForwardConstantPushInstruction(334) || hook.hookForwardConstantPushInstruction(-333)) {
                                hook.index = 0;
                                if (hook.hookForwardConstantPushInstruction(15)) {
                                    // Count
                                    Instruction i = hook.instructions[hook.index + 1];
                                    if (i instanceof GETSTATIC) {
                                        FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index + 1];
                                        addMethodToReturnField(client, "getMenuOptionsCount", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                    } else {
                                        FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index - 1];
                                        addMethodToReturnField(client, "getMenuOptionsCount", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                    }



                                    hook.index = 0;
                                    if(hook.hookForwardLDC(lookup)) {
                                        System.out.println("YAY!");
                                        int mark = -1;
                                        while(hook.seekForwardLocalVariableInstruction()) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];
                                            if(lvi instanceof StoreInstruction) {
                                                mark = lvi.getIndex();
                                                break;
                                            }
                                        }

                                        while(hook.hookForwardLocalVariableInstruction(mark)) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];

                                            if(lvi instanceof LoadInstruction) {
                                                i = hook.instructions[hook.index+1];
                                                if(i instanceof PUTSTATIC) {
                                                    FieldInstruction fi = (FieldInstruction) i;
                                                    addMethodToReturnField(client, "getMenuWidth", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                    break;
                                                }
                                            }
                                        }
                                    }

                                    hook.index = 0;
                                    // Height
                                    if(hook.seekForwardFieldInstruction()) {
                                        FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                        addMethodToReturnField(client, "getMenuHeight", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                    }

                                    // Draw?
                                    hook.index = 0;
                                    while (hook.seekForwardFieldInstruction()) {
                                        i = hook.instructions[hook.index];
                                        if (i instanceof PUTSTATIC) {
                                            FieldInstruction fi = (FieldInstruction) i;
                                            if (fi.getType(cpg).getType() == Constants.T_BOOLEAN) {
                                                addMethodToReturnField(client, "isMenuOpen", Type.getReturnType("Z"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                break;
                                            }
                                        }
                                    }

                                    // Type
                                    hook.index = 0;
                                    while(hook.hookForwardConstantPushInstruction(2)) {
                                        i = hook.instructions[hook.index+1];
                                        if (i instanceof PUTSTATIC) {
                                            FieldInstruction fi = (FieldInstruction) i;
                                            if (fi.getType(cpg).getType() == Constants.T_INT) {
                                                addMethodToReturnField(client, "getMenuType", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                break;
                                            }
                                        }
                                    }


                                    // X
                                    hook.index = 0;
                                    if(hook.hookForwardConstantPushInstruction(334) || hook.hookForwardConstantPushInstruction(-334)) {
                                        int mark = -1;
                                        while(hook.seekForwardLocalVariableInstruction()) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];
                                            if(lvi instanceof StoreInstruction) {
                                                mark = lvi.getIndex();
                                                break;
                                            }
                                        }

                                        while(hook.hookForwardLocalVariableInstruction(mark)) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];

                                            if(lvi instanceof LoadInstruction) {
                                                i = hook.instructions[hook.index+1];
                                                if(i instanceof PUTSTATIC) {
                                                    FieldInstruction fi = (FieldInstruction) i;
                                                    addMethodToReturnField(client, "getMenuY", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                    break;
                                                }
                                            }
                                        }
                                    }

                                    // Y
                                    hook.index = 0;
                                    if(hook.hookForwardConstantPushInstruction(512)) {
                                        int mark = -1;
                                        while(hook.seekForwardLocalVariableInstruction()) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];
                                            if(lvi instanceof StoreInstruction) {
                                                mark = lvi.getIndex();
                                                break;
                                            }
                                        }

                                        while(hook.hookForwardLocalVariableInstruction(mark)) {
                                            LocalVariableInstruction lvi = (LocalVariableInstruction)hook.instructions[hook.index];

                                            if(lvi instanceof LoadInstruction) {
                                                i = hook.instructions[hook.index+1];
                                                if(i instanceof PUTSTATIC) {
                                                    FieldInstruction fi = (FieldInstruction) i;
                                                    addMethodToReturnField(client, "getMenuX", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                    break;
                                                }
                                            }
                                        }
                                    }


                                }
                            }
                        }
                    }
                }
            }
        }
    }

    public void hackArrayCopy() {
        for (ClassGen cg : classes) {
            Method methods[] = cg.getMethods();
            if (methods.length == 1) {
                Method method = methods[0];
                if (method.isStatic() && method.isFinal() && method.getReturnType().getType() == Constants.T_VOID) {
                    String methodname = method.getName();
                    String methodclass = cg.getClassName();
                    String methodsig = method.getSignature();

                    for (ClassGen c : classes) {
                        ConstantPoolGen cpg = c.getConstantPool();
                        int find = cpg.lookupMethodref(methodclass, methodname, methodsig);
                        if (find != -1) {
                            cpg.setConstant(find, new ConstantMethodref(cpg.addClass("java/lang/System"), cpg.addNameAndType("arraycopy", "(Ljava/lang/Object;ILjava/lang/Object;II)V")));
                        }
                    }

                    return;
                }
            }
        }
    }

    public void hackMouse() {
        for (ClassGen cg : classes) {
            String interfaces[] = cg.getInterfaceNames();
            for (String iface : interfaces) {
                if (iface.indexOf("Mouse") != -1) {
                    classnames.put("Mouse", cg.getClassName());
                    ConstantPoolGen cpg = cg.getConstantPool();
                    cpg.setConstant(cpg.lookupUtf8("mouseClicked"), new ConstantUtf8("_mouseClicked"));
                    cpg.setConstant(cpg.lookupUtf8("mousePressed"), new ConstantUtf8("_mousePressed"));
                    cpg.setConstant(cpg.lookupUtf8("mouseReleased"), new ConstantUtf8("_mouseReleased"));
                    cpg.setConstant(cpg.lookupUtf8("mouseEntered"), new ConstantUtf8("_mouseEntered"));
                    cpg.setConstant(cpg.lookupUtf8("mouseExited"), new ConstantUtf8("_mouseExited"));
                    cpg.setConstant(cpg.lookupUtf8("mouseDragged"), new ConstantUtf8("_mouseDragged"));
                    cpg.setConstant(cpg.lookupUtf8("mouseMoved"), new ConstantUtf8("_mouseMoved"));
                    cpg.setConstant(cpg.lookupUtf8("focusGained"), new ConstantUtf8("_focusGained"));
                    cpg.setConstant(cpg.lookupUtf8("focusLost"), new ConstantUtf8("_focusLost"));
                    ((ConstantClass) cpg.getConstant(cg.getSuperclassNameIndex())).setNameIndex(cpg.addUtf8("com/bawt/hook/Mouse"));

                    return;
                }
            }
        }
    }

    public void hackKeyBoard() {
        for (ClassGen cg : classes) {
            String interfaces[] = cg.getInterfaceNames();
            for (String iface : interfaces) {
                if (iface.indexOf("Key") != -1) {
                    classnames.put("Keyboard", cg.getClassName());
                    ConstantPoolGen cpg = cg.getConstantPool();
                    cpg.setConstant(cpg.lookupUtf8("keyTyped"), new ConstantUtf8("_keyTyped"));
                    cpg.setConstant(cpg.lookupUtf8("keyPressed"), new ConstantUtf8("_keyPressed"));
                    cpg.setConstant(cpg.lookupUtf8("keyReleased"), new ConstantUtf8("_keyReleased"));
                    cpg.setConstant(cpg.lookupUtf8("focusGained"), new ConstantUtf8("_focusGained"));
                    cpg.setConstant(cpg.lookupUtf8("focusLost"), new ConstantUtf8("_focusLost"));
                    ((ConstantClass) cpg.getConstant(cg.getSuperclassNameIndex())).setNameIndex(cpg.addUtf8("com/bawt/hook/Keyboard"));
                    return;
                }
            }
        }
    }

    public void hackCanvas() {
        for (ClassGen cg : classes) {
            if (cg.getSuperclassName().equals("java.awt.Canvas")) {
                ConstantPoolGen cpg = cg.getConstantPool();
                ((ConstantClass) cpg.getConstant(cg.getSuperclassNameIndex())).setNameIndex(cpg.addUtf8("com/bawt/hook/Canvas"));
                classnames.put("Canvas", cg.getClassName());
                return;
            }
        }
    }

    public boolean hookCanvas() {
        ClassGen client = findClass("client");
        for (ClassGen cg : classes) {
            Field fields[] = cg.getFields();
            for (Field f : fields) {
                if (f.isStatic()) {
                    if (f.getType().toString().equals("java.awt.Canvas")) {
                        addMethodToReturnField(client, "getCanvas", Type.getReturnType("Ljava/awt/Canvas;"), cg.getClassName(), f.getName(), f.getType(), Constants.GETSTATIC);
                        return true;
                    }
                }
            }
        }

        return false;
    }

    public boolean hookUserInput() {
        ClassGen client = findClass("client");
        String mouse = classnames.get("Mouse");
        String keyboard = classnames.get("Keyboard");
        boolean m = false;
        boolean k = false;

        for (ClassGen cg : classes) {
            Field fields[] = cg.getFields();
            for (Field f : fields) {
                if (f.isStatic()) {
                    String type = f.getType().toString();

                    if (!m && type.equals(mouse)) {
                        addMethodToReturnField(client, "getMouse", Type.getReturnType("Lcom/bawt/hook/Mouse;"), cg.getClassName(), f.getName(), f.getType(), Constants.GETSTATIC);
                        m = true;
                    } else if (!k && type.equals(keyboard)) {
                        addMethodToReturnField(client, "getKeyboard", Type.getReturnType("Lcom/bawt/hook/Keyboard;"), cg.getClassName(), f.getName(), f.getType(), Constants.GETSTATIC);
                        k = true;
                    }

                    if (m && k) {
                        return true;
                    }
                }
            }
        }

        return m && k;
    }


    // TODO Make this hook able to identify them 100% correctly. (2 can possibly get mixed);
    public boolean hookSkills() {
        ClassGen client = findClass("client");
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            int lookup = cpg.lookupFieldref(classnames.get("ExpMax"), fieldnames.get("ExpMax"), "[I");
            if (lookup != -1) {
                String classname = cg.getClassName();
                Method methods[] = cg.getMethods();
                for (Method m : methods) {
                    if (!m.isAbstract() && m.isStatic() && m.getReturnType().toString().equals("boolean")) {
                        MethodGen mg = new MethodGen(m, classname, cpg);
                        InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                        if (hook.hookForwardFieldInstruction(lookup)) {
                            if (hook.seekForwardFieldInstruction()) {
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                addMethodToReturnField(client, "getSkillMaxes", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                hook.seekReverseFieldInstruction();
                                hook.seekReverseFieldInstruction();
                                hook.seekReverseFieldInstruction();
                                fi = (FieldInstruction) hook.instructions[hook.index];
                                addMethodToReturnField(client, "getSkillLevels", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                fi = (FieldInstruction) hook.instructions[hook.index];
                                hook.seekReverseFieldInstruction();
                                fi = (FieldInstruction) hook.instructions[hook.index];
                                addMethodToReturnField(client, "getSkillExperiences", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                return true;
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    public boolean hookExpMaxes() {
        ClassGen client = findClass("client");
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            int pow = cpg.lookupMethodref("java.lang.Math", "pow", "(DD)D");
            if (pow != -1) {
                Method sinit = cg.containsMethod(Constants.STATIC_INITIALIZER_NAME, "()V");
                String classname = cg.getClassName();
                if(sinit != null) {
                    MethodGen mg = new MethodGen(sinit, classname, cpg);
                    InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                    if (hook.hookForwardInvokeInstruction(pow)) {
                        if (hook.seekForwardFieldInstruction()) {
                            FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                            addMethodToReturnField(client, "getSkillExperiencesMax", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                            classnames.put("ExpMax", fi.getClassName(cpg));
                            fieldnames.put("ExpMax", fi.getName(cpg));
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    public boolean hookCharacterCoordinates() {
        ClassGen character = findClass(classnames.get("Character"));
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            int lookupxloc = cpg.lookupFieldref(classnames.get("Character"), fieldnames.get("xloc"), "[I");
            int lookupyloc = cpg.lookupFieldref(classnames.get("Character"), fieldnames.get("yloc"), "[I");
            if (lookupxloc != -1 && lookupyloc != -1) {
                String classname = cg.getClassName();
                Method methods[] = cg.getMethods();
                for (Method m : methods) {
                    if (!m.isAbstract() && m.isStatic() && m.getSignature().indexOf("L" + classnames.get("Character") + ";") != -1) {
                        MethodGen mg = new MethodGen(m, classname, cpg);
                        InstructionList il = mg.getInstructionList();
                        InstructionHook hook = new InstructionHook(il.getInstructions());
//                        if (hook.hookForwardConstantPushInstruction(13184) || hook.hookForwardInvokeInstruction(-13185)) {
                            hook.index = 0;
                        int add = 0;
                            if (hook.hookForwardFieldInstruction(lookupxloc)) {
                                hook.seekForwardFieldInstruction();
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                while (!(fi instanceof PUTFIELD)) {
                                    hook.seekForwardFieldInstruction();
                                    fi = (FieldInstruction) hook.instructions[hook.index];
                                }
                                add++;
                                addMethodToReturnField(character, "getX", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETFIELD);
                            }
                            hook.index = 0;
                            if (hook.hookForwardFieldInstruction(lookupyloc)) {
                                hook.seekForwardFieldInstruction();
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                while (!(fi instanceof PUTFIELD)) {
                                    hook.seekForwardFieldInstruction();
                                    fi = (FieldInstruction) hook.instructions[hook.index];
                                }
                                add++;
                                addMethodToReturnField(character, "getY", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETFIELD);
                            }
                        if(add == 2) return true;
//                        }
                    }
                }
            }
        }

        return false;
    }

    public boolean hookChatStates() {
        ClassGen client = findClass("client");
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            String classname = cg.getClassName();
            Method methods[] = cg.getMethods();
            for (Method m : methods) {
                if (!m.isAbstract() && m.isStatic()) {
                    MethodGen mg = new MethodGen(m, classname, cpg);
                    InstructionList il = mg.getInstructionList();
                    InstructionFinder finder = new InstructionFinder(il);
                    Iterator iterator = finder.search("ConstantPushInstruction IREM", new InstructionFinder.CodeConstraint() {
                        public boolean checkCode(InstructionHandle[] matches) {
                            ConstantPushInstruction cpi = (ConstantPushInstruction) matches[0].getInstruction();
                            return cpi.getValue().intValue() == 4;
                        }
                    });

                    if (iterator.hasNext()) {
                        InstructionHook hook = new InstructionHook(il.getInstructions());
                        hook.index = ((InstructionHandle[]) iterator.next())[0].getPosition();
                        hook.seekForwardConstantPushInstruction();

                        hook.seekForwardFieldInstruction();
                        FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                        while (!(fi instanceof GETSTATIC) || !(fi.getFieldType(cpg).toString().equals("int"))) {
                            hook.seekForwardFieldInstruction();
                            fi = (FieldInstruction) hook.instructions[hook.index];
                        }
                        addMethodToReturnField(client, "getChatStatePublic", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);

                        hook.seekForwardFieldInstruction();
                        fi = (FieldInstruction) hook.instructions[hook.index];
                        while (!(fi instanceof GETSTATIC) || !(fi.getFieldType(cpg).toString().equals("int"))) {
                            hook.seekForwardFieldInstruction();
                            fi = (FieldInstruction) hook.instructions[hook.index];
                        }
                        addMethodToReturnField(client, "getChatStatePrivate", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);

                        hook.seekForwardFieldInstruction();
                        fi = (FieldInstruction) hook.instructions[hook.index];
                        while (!(fi instanceof GETSTATIC) || !(fi.getFieldType(cpg).toString().equals("int"))) {
                            hook.seekForwardFieldInstruction();
                            fi = (FieldInstruction) hook.instructions[hook.index];
                        }
                        addMethodToReturnField(client, "getChatStateRequest", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public boolean hookCharacterHpInfo() {
        String character = classnames.get("Character");
        ClassGen chr = findClass(character);
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            int lookup = cpg.lookupInteger(0x80ff80);
            if (lookup != -1) {
                lookup = cpg.lookupInteger(0x00FF00);
                if (lookup != -1) {
                    String classname = cg.getClassName();
                    Method methods[] = cg.getMethods();
                    for (Method m : methods) {
                        if (m.isStatic()) {
                            MethodGen mg = new MethodGen(m, classname, cpg);
                            InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                            while (hook.hookForwardConstantPushInstruction(30)) {
                                Instruction inst1 = hook.instructions[hook.index + 1];
                                Instruction inst3 = hook.instructions[hook.index + 3];
                                if (inst1 instanceof IMUL) {
                                    FieldInstruction hpcur = (FieldInstruction) hook.instructions[hook.index - 1];
                                    addMethodToReturnField(chr, "getHp", Type.getReturnType("I"), hpcur.getClassName(cpg), hpcur.getName(cpg), hpcur.getType(cpg), Constants.GETFIELD);
                                    FieldInstruction hpmax = (FieldInstruction) hook.instructions[hook.index + 3];
                                    addMethodToReturnField(chr, "getHpMax", Type.getReturnType("I"), hpmax.getClassName(cpg), hpmax.getName(cpg), hpmax.getType(cpg), Constants.GETFIELD);
                                    hook.seekReverseFieldInstruction();
                                    hook.seekReverseFieldInstruction();
                                    hook.seekReverseFieldInstruction();
                                    FieldInstruction height = (FieldInstruction) hook.instructions[hook.index];
                                    addMethodToReturnField(chr, "getHeight", Type.getReturnType("I"), height.getClassName(cpg), height.getName(cpg), height.getType(cpg), Constants.GETFIELD);
//                                    hook.seekForwardInvokeInstruction();
//                                    InvokeInstruction chr2scr = (InvokeInstruction) hook.instructions[hook.index];
//                                    classnames.put("chr2scrclass", chr2scr.getClassName(cpg));
//                                    classnames.put("chr2scrname", chr2scr.getMethodName(cpg));
//                                    classnames.put("chr2scrsig", chr2scr.getSignature(cpg));
                                    return true;
                                } else if (inst3 instanceof IMUL) {
                                    FieldInstruction hpcur = (FieldInstruction) hook.instructions[hook.index + 2];
                                    addMethodToReturnField(chr, "getHp", Type.getReturnType("I"), hpcur.getClassName(cpg), hpcur.getName(cpg), hpcur.getType(cpg), Constants.GETFIELD);
                                    FieldInstruction hpmax = (FieldInstruction) hook.instructions[hook.index + 5];
                                    addMethodToReturnField(chr, "getHpMax", Type.getReturnType("I"), hpmax.getClassName(cpg), hpmax.getName(cpg), hpmax.getType(cpg), Constants.GETFIELD);
                                    hook.seekReverseFieldInstruction();
                                    hook.seekReverseFieldInstruction();
                                    FieldInstruction height = (FieldInstruction) hook.instructions[hook.index];
                                    addMethodToReturnField(chr, "getHeight", Type.getReturnType("I"), height.getClassName(cpg), height.getName(cpg), height.getType(cpg), Constants.GETFIELD);
//                                    hook.seekForwardInvokeInstruction();
//                                    InvokeInstruction chr2scr = (InvokeInstruction) hook.instructions[hook.index];
//                                    classnames.put("chr2scrclass", chr2scr.getClassName(cpg));
//                                    classnames.put("chr2scrname", chr2scr.getMethodName(cpg));
//                                    classnames.put("chr2scrsig", chr2scr.getSignature(cpg));
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    public boolean hookWorldCoords() {
        ClassGen client = findClass("client");
        ClassGen character = findClass(classnames.get("Character"));
        for (ClassGen cg : classes) {
            ConstantPoolGen cpg = cg.getConstantPool();
            int lookup = cpg.lookupString("T2 - ");
            if (lookup != -1) {
                String classname = cg.getClassName();
                Method methods[] = cg.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        MethodGen mg = new MethodGen(m, classname, cpg);
                        InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                        if (hook.hookForwardLDC(lookup)) {
                            for (int i = 0; i < 4; i++) {
                                if (!hook.seekForwardFieldInstruction()) {
                                    return false;
                                }
                            }
                            boolean x = false;
                            boolean xloc = false;
                            boolean y = false;
                            boolean yloc = false;
                            for (int i = 0; i < 3; i++) {
                                hook.seekForwardFieldInstruction();
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                String type = fi.getType(cpg).toString();
                                if (type.equals("int") && !x) {
                                    addMethodToReturnField(client, "getWorldX", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                    x = true;
                                } else if (type.equals("int[]")) {
                                    addMethodToReturnField(character, "getXloc", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETFIELD);
                                    fieldnames.put("xloc", fi.getFieldName(cpg));
                                    xloc = true;
                                }
                                if (x && xloc) {
                                    break;
                                }
                            }
                            for (int i = 0; i < 5; i++) {
                                hook.seekForwardFieldInstruction();
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                String type = fi.getType(cpg).toString();
                                if (type.equals("int")) {
                                    addMethodToReturnField(client, "getWorldY", Type.getReturnType("I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                    y = true;
                                } else if (type.equals("int[]")) {
                                    addMethodToReturnField(character, "getYloc", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETFIELD);
                                    fieldnames.put("yloc", fi.getFieldName(cpg));
                                    yloc = true;
                                }
                                if (y && yloc) {
                                    break;
                                }
                            }
                            return x && xloc && y && yloc;
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookChatArrays() {
        Member close = findString("@cr1@");
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupFieldref(close.classname, close.name, "L" + close.signature + ";");
            if (lookup != -1) {
                String classname = c.getClassName();
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        if (m.getName().indexOf("init") == -1) {
                            MethodGen mg = new MethodGen(m, classname, cpg);
                            InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                            if (hook.hookForwardFieldInstruction(lookup)) {
                                Instruction inst = hook.instructions[hook.index];
                                if (inst instanceof GETSTATIC) {
                                    String sw = classnames.get("StringWrapper");
                                    int count = 0;
                                    ClassGen client = findClass("client");
                                    int select = 0;
                                    boolean check = false;
                                    while (hook.seekReverseFieldInstruction()) {
                                        FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                        Type fieldtype = fi.getFieldType(cpg);
                                        if (fieldtype.toString().equals(sw + "[]")) {
                                            if (select == 0) {
                                                addMethodToReturnField(client, "getChatMessages", Type.getReturnType("[Lcom/bawt/hook/StringWrapper;"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                select++;
                                            } else if (select == 1) {
                                                addMethodToReturnField(client, "getChatNames", Type.getReturnType("[Lcom/bawt/hook/StringWrapper;"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                                select++;
                                            }
                                        } else if (fieldtype.toString().equals("int[]")) {
                                            addMethodToReturnField(client, "getChatTypes", Type.getReturnType("[I"), fi.getClassName(cpg), fi.getName(cpg), fi.getType(cpg), Constants.GETSTATIC);
                                            check = true;
                                        }
                                        if (check && select == 2)
                                            return true;
                                        count++;
                                        if (count > 10) {
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookMyPlayer() {
        String player = classnames.get("Player");
        for (ClassGen c : classes) {
            Field fields[] = c.getFields();
            for (Field f : fields) {
                if (f.getType().toString().equals(player)) {
                    ClassGen client = findClass("client");
                    addMethodToReturnField(client, "getPlayerMe", Type.getType(Player.class), c.getClassName(), f.getName(), f.getType(), Constants.GETSTATIC);
                    return true;
                }
            }
        }
        return false;
    }

    private boolean hookLevels() {
        Member level = findString("level)2");
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupFieldref(level.classname, level.name, 'L' + level.signature + ';');
            if (lookup != -1) {
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        if (m.getSignature().indexOf(classnames.get("NPC")) != -1) {
                            String classname = c.getClassName();
                            MethodGen mg = new MethodGen(m, classname, cpg);
                            InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                            if (hook.hookForwardFieldInstruction(lookup)) {
                                String npc = classnames.get("NPC");
                                String player = classnames.get("Player");
                                boolean np = false;
                                boolean pl = false;
                                while (hook.seekReverseFieldInstruction()) {
                                    FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                    Member check = new Member(fi.getClassName(cpg), fi.getFieldName(cpg), fi.getFieldType(cpg).toString());
                                    if (check.signature.equals("int")) {
                                        if (check.classname.equals(npc) && !np) {
                                            np = true;
                                            addMethodToReturnField(findClass(npc), "getLevel", Type.INT, npc, fi.getFieldName(cpg), fi.getFieldType(cpg), Constants.GETFIELD);
                                        } else if (check.classname.equals(player) && !pl) {
                                            pl = true;
                                            addMethodToReturnField(findClass(player), "getLevel", Type.INT, player, fi.getFieldName(cpg), fi.getFieldType(cpg), Constants.GETFIELD);
                                        }
                                    }
                                    if (np && pl)
                                        return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private Member findString(String str) {
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupString(str);
            if (lookup != -1) {
                String classname = c.getClassName();
                Method clinit = c.containsMethod(Constants.STATIC_INITIALIZER_NAME, "()V");
                MethodGen mg = new MethodGen(clinit, classname, cpg);
                InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                hook.hookForwardLDC(lookup);
                hook.seekForwardFieldInstruction();
                PUTSTATIC ps = (PUTSTATIC) hook.instructions[hook.index];
                Member m = new Member(ps.getClassName(cpg), ps.getFieldName(cpg), ps.getFieldType(cpg).toString());
                while (hook.seekForwardFieldInstruction()) {
                    FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                    Member check = new Member(fi.getClassName(cpg), fi.getFieldName(cpg), fi.getFieldType(cpg).toString());
                    if (check.equals(m)) {
                        if (hook.seekForwardFieldInstruction()) {
                            ps = (PUTSTATIC) hook.instructions[hook.index];
                            m = new Member(ps.getClassName(cpg), ps.getFieldName(cpg), ps.getFieldType(cpg).toString());
                        }
                    }
                }
                return m;
            }
        }
        return null;
    }

    private boolean hookMonsterID() {
        String npcname = classnames.get("NPC");
        ClassGen npc = findClass(npcname);
        for (ClassGen c : classes) {
            Method methods[] = c.getMethods();
            for (Method m : methods) {
                if (m.isStatic()) {
                    if (m.getReturnType().toString().equals(npcname)) {
                        ConstantPoolGen cpg = c.getConstantPool();
                        MethodGen mg = new MethodGen(m, c.getClassName(), cpg);
                        Instruction[] list = mg.getInstructionList().getInstructions();
                        for (int i = list.length - 1; i > -1; i--) {
                            Instruction inst = list[i];
                            if (inst instanceof PUTFIELD) {
                                PUTFIELD pf = (PUTFIELD) inst;
                                addMethodToReturnField(npc, "getID", Type.INT, npcname, pf.getFieldName(cpg), pf.getType(cpg), Constants.GETFIELD);
                                return true;
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookMonsterComposite() {
        ClassGen monster = findClass(classnames.get("Monster"));
        monster.addInterface("com.bawt.hook.Monster");
        Field fields[] = monster.getFields();
        int a = 0;
        for (Field f : fields) {
            if (!f.isStatic()) {
                addMethodToReturnField(monster, "getComposite", Type.getType(NPC.class), classnames.get("Monster"), f.getName(), f.getType(), Constants.GETFIELD);
                classnames.put("NPC", f.getType().toString());
                a++;
                break;
            }
        }

        ClassGen npc = findClass(classnames.get("NPC"));
        npc.addInterface("com.bawt.hook.NPC");
        String sw = classnames.get("StringWrapper");
        fields = npc.getFields();
        for (Field f : fields) {
            if (!f.isStatic()) {
                if (f.getSignature().equals("L" + sw + ";")) {
                    addMethodToReturnField(npc, "getName", Type.getType(StringWrapper.class), classnames.get("NPC"), f.getName(), f.getType(), Constants.GETFIELD);
                    a++;
                    break;
                }
            }
        }
        return a == 2;
    }

    private boolean hookPlayerCharacter() {
        int a = 0;
        ClassGen player = findClass(classnames.get("Player"));
        player.addInterface("com.bawt.hook.Player");
        String sw = classnames.get("StringWrapper");
        Field fields[] = player.getFields();
        for (Field f : fields) {
            if (!f.isStatic()) {
                if (f.getSignature().equals("L" + sw + ";")) {
                    addMethodToReturnField(player, "getName", Type.getType(StringWrapper.class), classnames.get("Player"), f.getName(), f.getType(), Constants.GETFIELD);
                    a++;
                    break;
                }
            }
        }
        classnames.put("Character", player.getSuperclassName());
        ClassGen character = findClass(classnames.get("Character"));
        character.addInterface("com.bawt.hook.Character");
        fields = character.getFields();
        for (Field f : fields) {
            if (!f.isStatic()) {
                if (f.getSignature().equals("L" + sw + ";")) {
                    addMethodToReturnField(character, "getMessage", Type.getType(StringWrapper.class), classnames.get("Character"), f.getName(), f.getType(), Constants.GETFIELD);
                    a++;
                    break;
                }
            }
        }
        return a == 2;
    }

    private boolean hookMonsterArrays() {
        ClassGen client = findClass("client");
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupString("gnp2 pos:");
            if (lookup != -1) {
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        MethodGen mg = new MethodGen(m, c.getClassName(), cpg);
                        InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                        if (hook.hookForwardLDC(lookup)) {
                            int count = 0;
                            while (hook.seekReverseFieldInstruction()) {
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                Type ftype = fi.getFieldType(cpg);
                                count++;
                                if (count == 1) {
                                    addMethodToReturnField(client, "getMonsterValids", ftype, fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                } else if (count == 2) {
                                    classnames.put("Monster", ftype.toString().replaceAll("\\[\\]", ""));
                                    addMethodToReturnField(client, "getMonsters", Type.getReturnType("[Lcom/bawt/hook/Monster;"), fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                } else {
                                    addMethodToReturnField(client, "getMonsterCount", Type.INT, fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                }
                                if (count == 3) {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookPlayerArrays() {
        ClassGen client = findClass("client");
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupString("gpp2 pos:");
            if (lookup != -1) {
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        MethodGen mg = new MethodGen(m, c.getClassName(), cpg);
                        InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                        if (hook.hookForwardLDC(lookup)) {
                            int count = 0;
                            while (hook.seekReverseFieldInstruction()) {
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                Type ftype = fi.getFieldType(cpg);
                                count++;
                                if (count == 1) {
                                    addMethodToReturnField(client, "getPlayerValids", ftype, fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                } else if (count == 2) {
                                    classnames.put("Player", ftype.toString().replaceAll("\\[\\]", ""));
                                    addMethodToReturnField(client, "getPlayers", Type.getReturnType("[Lcom/bawt/hook/Player;"), fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                } else {
                                    addMethodToReturnField(client, "getPlayerCount", Type.INT, fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                }
                                if (count == 3) {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookUserPass() {
        ClassGen client = findClass("client");
        Member rsa = findRsaEnc();
        for (ClassGen c : classes) {
            ConstantPoolGen cpg = c.getConstantPool();
            int lookup = cpg.lookupMethodref(rsa.classname, rsa.name, rsa.signature);
            if (lookup != -1) {
                Method methods[] = c.getMethods();
                for (Method m : methods) {
                    if (m.isStatic()) {
                        MethodGen mg = new MethodGen(m, c.getClassName(), cpg);
                        InstructionHook hook = new InstructionHook(mg.getInstructionList().getInstructions());
                        if (hook.hookForwardInvokeInstruction(lookup)) {
                            String sw = classnames.get("StringWrapper");
                            int count = 0;
                            while (hook.seekReverseFieldInstruction()) {
                                FieldInstruction fi = (FieldInstruction) hook.instructions[hook.index];
                                Type ftype = fi.getFieldType(cpg);
                                if (ftype.toString().equals(sw)) {
                                    count++;
                                    if (count == 1) {
                                        addMethodToReturnField(client, "getPassword", Type.getType(StringWrapper.class), fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                    } else {
                                        addMethodToReturnField(client, "getUsername", Type.getType(StringWrapper.class), fi.getClassName(cpg), fi.getFieldName(cpg), ftype, Constants.GETSTATIC);
                                    }
                                }

                                if (count == 2) {
                                    return true;
                                }
                            }
                        }
                    }
                }
            }
        }
        return false;
    }

    private boolean hookStringWrapper() {
        for (ClassGen c : classes) {
            String interfaces[] = c.getInterfaceNames();
            for (String iface : interfaces) {
                ClassGen find = findClass(iface);
                if (find != null) {
                    if (find.getMethods().length == 0 && find.getFields().length == 0) {
                        c.addInterface("com.bawt.hook.StringWrapper");
                        Field fields[] = c.getFields();
                        for (Field f : fields) {
                            if (!f.isStatic()) {
                                String signature = f.getSignature();
                                if (signature.equals("[B")) {
                                    classnames.put("StringWrapper", c.getClassName());
                                    addMethodToReturnField(c, "getBytes", f.getType(), classnames.get("StringWrapper"), f.getName(), f.getType(), Constants.GETFIELD);
                                    return true;
                                }
                            }
                        }
                        break;
                    }
                }
            }
        }
        return false;
    }

    private ClassGen addMethodToReturnField(ClassGen cg, String methodname, Type methodtype, String classname, String fieldname, Type fieldtype, short type) {
        ConstantPoolGen cpg = cg.getConstantPool();
        InstructionList il = new InstructionList();
        InstructionFactory factory = new InstructionFactory(cg, cpg);
        MethodGen mg = new MethodGen(Constants.ACC_PUBLIC, methodtype, Type.NO_ARGS, new String[]{}, methodname, cg.getClassName(), il, cpg);
        il.append(new ALOAD(0));
        il.append(factory.createFieldAccess(classname, fieldname, fieldtype, type));
        il.append(InstructionFactory.createReturn(fieldtype));
        mg.setMaxLocals();
        mg.setMaxStack();
        cg.addMethod(mg.getMethod());
        System.out.println(methodname + "() added to class \"" + cg.getClassName() + "\" gets " + ((type == Constants.GETSTATIC) ? "static " : "") + "field \"" + classname + '.' + fieldtype.toString() + ' ' + fieldname + '\"');
        return cg;
    }

    private Member findRsaEnc() {
        for (ClassGen c : classes) {
            Method methods[] = c.getMethods();
            for (Method m : methods) {
                String sig = m.getSignature();
                if (sig.indexOf("BigInteger") != -1) {
                    return new Member(c.getClassName(), m.getName(), sig);
                }
            }
        }
        return null;
    }

    public ClassGen findClass(String name) {
        for (ClassGen c : classes) {
            if (c.getClassName().equals(name)) {
                return c;
            }
        }
        return null;
    }
}