package com.bawt.update;

import org.apache.bcel.generic.*;

public class InstructionHook {
    public Instruction instructions[];
    public int index;

    public InstructionHook(Instruction[] i) {
        index = -1;
        instructions = i;
    }

    public boolean hookForwardFieldInstruction(int ref) {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof FieldInstruction) {
                FieldInstruction fs = (FieldInstruction) instruction;
                if (fs.getIndex() == ref) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean hookForwardConstantPushInstruction(int con) {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof ConstantPushInstruction) {
                ConstantPushInstruction cpi = (ConstantPushInstruction) instruction;
                if (cpi.getValue().intValue() == con) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean hookReverseConstantPushInstruction(int con) {
        for (int i = index - 1; i > -1; i--) {
            Instruction instruction = instructions[i];
            if (instruction instanceof ConstantPushInstruction) {
                ConstantPushInstruction cpi = (ConstantPushInstruction) instruction;
                if (cpi.getValue().intValue() == con) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean hookForwardInvokeInstruction(int ref) {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof InvokeInstruction) {
                InvokeInstruction invoke = (InvokeInstruction) instruction;
                if (invoke.getIndex() == ref) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean hookForwardLocalVariableInstruction(int ref) {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof LocalVariableInstruction) {
                LocalVariableInstruction lvi = (LocalVariableInstruction) instruction;
                if (lvi.getIndex() == ref) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean hookForwardLDC(int ref) {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof LDC) {
                LDC ldc = (LDC) instruction;
                if (ldc.getIndex() == ref) {
                    index = i;
                    return true;
                }
            }
        }
        return false;
    }

    public boolean seekReverseFieldInstruction() {
        for (int i = index - 1; i > -1; i--) {
            Instruction instruction = instructions[i];
            if (instruction instanceof FieldInstruction) {
                index = i;
                return true;
            }
        }
        return false;
    }

    public boolean seekForwardFieldInstruction() {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof FieldInstruction) {
                index = i;
                return true;
            }
        }
        return false;
    }

    public boolean seekForwardInvokeInstruction() {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof InvokeInstruction) {
                index = i;
                return true;
            }
        }
        return false;
    }

    public boolean seekForwardConstantPushInstruction() {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof ConstantPushInstruction) {
                index = i;
                return true;
            }
        }
        return false;
    }
    public boolean seekForwardLocalVariableInstruction() {
        for (int i = index + 1; i < instructions.length; i++) {
            Instruction instruction = instructions[i];
            if (instruction instanceof LocalVariableInstruction) {
                index = i;
                return true;
            }
        }
        return false;
    }
}
